import pytest
from aeromesh.structs.structures import WindFarm, Domain
from aeromesh.geometry.functions2D import *
from aeromesh.terrain.terrain import buildTerrain2D

@pytest.fixture
def domain():
    d = Domain()
    d.setDomain([-1200, 1200], [-1200, 1200], [0, 1])
    return d

def test_turbinegen_full(domain):
    gmsh.model.add("2D Turbine Generation Test")

    params = dict()
    params['domain'] = {
        'x_range': [-1200, 1200],
        'y_range': [-1200, 1200],
        'inflow_angle': 0
    }
    params['refine'] = {
        'background_length_scale': 100,
        'turbine': {
            'type': 'wake'
        },
        'farm': {}
    }

    params['refine']['turbine']['num_turbines'] = 2
    params['refine']['turbine']['length_scale'] = 30
    params['refine']['background_length_scale'] = 100
    params['refine']['farm']['length_scale'] = 70
    params['refine']['turbine']['threshold_upstream_distance'] = 240
    params['refine']['turbine']['threshold_downstream_distance'] = 300
    params['refine']['turbine']['threshold_rotor_distance'] = 100
    params['refine']['turbine']['shudder'] = 50

    params['refine']['turbine'][1] = {
        'x': 200,
        'y': 400,
    }
    params['refine']['turbine'][2] = {
        'x': -800,
        'y': 300,
    }

    struct = buildTerrain2D(params, domain)
    gmsh.model.geo.addPlaneSurface([struct], tag=999)
    wf = WindFarm()

    turbines = buildFarms2D(params, wf, domain)

    gmsh.model.remove()

    assert len(turbines) == 2


import gmsh
import sys, os
import yaml
import aeromesh.structs as structs
import aeromesh.terrain as tr
import aeromesh.geometry as geometry
import meshio
import numpy as np
import math
import tempfile

def toXDMF(ndim, name):
    tempdir = tempfile.TemporaryDirectory()
    gmsh.write(os.path.join(tempdir.name, "dummy.msh"))

    msh = meshio.read(os.path.join(tempdir.name, "dummy.msh"))

    def create_mesh(mesh, type, prune_dim=False):
        cells = mesh.get_cells_type(type)
        cell_data = mesh.get_cell_data("gmsh:physical", type)
        points = mesh.points if prune_dim == False else mesh.points[:, :2]
        out_mesh = meshio.Mesh(points=points, cells={type: cells}, cell_data={"facet_tags": [cell_data.astype(np.int32)]})
        return out_mesh

    if ndim == 3:
        tetras = create_mesh(msh, 'tetra')
        tris = create_mesh(msh, 'triangle')
        meshio.write(name + '.xdmf', tetras)
        meshio.write(name + '_boundary.xdmf', tris)

    else:
        tris = create_mesh(msh, 'triangle', prune_dim=True)
        lines = create_mesh(msh, 'line', prune_dim=True)
        meshio.write(name + '.xdmf', tris)
        meshio.write(name + '_boundary.xdmf', lines)

    tempdir.cleanup()

def generate2DMesh(params):

    """
    Executes a 2D meshing operation on params. Outputs results in the file out.vtk.

    :param params: A dictionary of the required parameters.
    :type params: dict()

    """
    gmsh.model.add("User Model")

    scale = params['refine']['global_scale']
    params['refine']['turbine']['length_scale'] *= scale
    params['refine']['farm']['length_scale'] *= scale
    params['refine']['background_length_scale'] *= scale
    farmType = params['domain']['type']

    domain = structs.Domain()

    wf = structs.WindFarm()
    if farmType == 'box':
        farmBorder = tr.buildTerrain2D(params, domain)
    elif farmType == 'cylinder':
        farmBorder = tr.buildTerrainCircle(params, domain)
    else:
        raise Exception("Invalid farm type specified. Farm types must be in [box, cylinder].")
    
    farm = gmsh.model.geo.addPlaneSurface([farmBorder], tag=999)
    sp1 = gmsh.model.geo.addPhysicalGroup(2, [farm], tag=0)

    fields = [999]
    fields.extend(geometry.buildFarms2D(params, wf, domain))
    fields.extend(geometry.generateCustomRefines(params))
    if params['refine']['farm']['type'] != 'none':
        fields.append(998)
        geometry.refineFarm2D(params, wf)

    mesher = gmsh.model.mesh.field.add("Min")
    gmsh.model.mesh.field.setNumbers(mesher, "FieldsList", fields)
    gmsh.model.mesh.field.setAsBackgroundMesh(mesher)

    gmsh.model.geo.synchronize()

    gmsh.model.mesh.removeDuplicateNodes()
    gmsh.model.mesh.removeDuplicateElements()

    gmsh.model.mesh.generate(2)

def generate3DMesh(params):

    """
    Executes a 3D meshing operation on params. Outputs results in the file out.vtk.

    :param params: A dictionary of the required parameters.
    :type params: dict()

    """

    gmsh.model.add("User Model")

    scale = params['refine']['global_scale']
    farmType = params['domain']['type']
    params['refine']['turbine']['length_scale'] *= scale
    params['refine']['farm']['length_scale'] *= scale
    params['refine']['background_length_scale'] *= scale

    domain = structs.Domain()
    terrain = None
    if farmType == 'box':
        try:
            terrainDefined = params['domain']['terrain_path']
            terrain = tr.buildTerrainFromFile(params, domain)
        except KeyError:
            terrain = tr.buildTerrainDefault(params, domain)
    elif farmType == 'cylinder':
        terrain = tr.buildTerrainCylinder(params, domain)
    else:
        raise Exception("Invalid farm type specified. Farm types must be in [box, cylinder].")

    if farmType == 'box':
        v1 = gmsh.model.geo.addVolume([terrain], tag=1)
        vp1 = gmsh.model.geo.addPhysicalGroup(3, [v1], tag=0)
        
    wf = structs.WindFarm()
    fields = geometry.generateTurbines(params, domain, wf)
    fields.append(999) #Background field, reserved number
    farmRefine = geometry.refineFarm3D(params, wf)
    if farmRefine:
        fields.append(farmRefine)
    fields.extend(geometry.generateCustomRefines(params))

    gmsh.model.geo.synchronize()
    mesher = gmsh.model.mesh.field.add("Min")
    gmsh.model.mesh.field.setNumbers(mesher, "FieldsList", fields)
    gmsh.model.mesh.field.setAsBackgroundMesh(mesher)

    gmsh.model.mesh.removeDuplicateNodes()
    gmsh.model.mesh.removeDuplicateElements()

    gmsh.model.mesh.generate(3)

    geometry.anisotropyScale(params)

    if farmType == 'cylinder' and domain.interp:
        geometry.cylinderTerrainAdjustment(domain, params)

def main():
    gmsh.initialize()

    if len(sys.argv) < 2:
        raise Exception("Input file not specified.")
    filename = sys.argv[1]
    with open(filename) as input:
        params = yaml.safe_load(input)

    setYAMLDefaults(params)
    verifyYAML(params)

    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
    gmsh.option.setNumber("Mesh.OptimizeThreshold", 1)
    if params['suppress_out'] > 0:
        gmsh.option.setNumber("General.Verbosity", 0)

    params['domain']['inflow_angle'] *= math.pi / 180

    if params['domain']['dimension'] == 3:
        generate3DMesh(params)
    else:
        generate2DMesh(params)
    gmsh.model.mesh.optimize()

    filename = params['filename']
    
    if params['filetype'] != 'xdmf':
        filename = filename + params['filetype']
        gmsh.write(filename)
    else:
        ndim = params['domain']['dimension']
        toXDMF(ndim, filename)

    if '-v' in sys.argv:
        gmsh.fltk.run()

    gmsh.finalize()

def runAeroMesh(params):
    gmsh.initialize()

    setYAMLDefaults(params)
    verifyYAML(params)

    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
    gmsh.option.setNumber("Mesh.OptimizeThreshold", 1)
    if params['suppress_out'] > 0:
        gmsh.option.setNumber("General.Verbosity", 0)

    params['domain']['inflow_angle'] *= math.pi / 180

    if params['domain']['dimension'] == 3:
        generate3DMesh(params)
    else:
        generate2DMesh(params)
    gmsh.model.mesh.optimize()

    filename = params['filename']
    
    if params['filetype'] != 'xdmf':
        filename = filename + params['filetype']
        gmsh.write(filename)
    else:
        ndim = params['domain']['dimension']
        toXDMF(ndim, filename)

    gmsh.finalize()

def setYAMLDefaults(params):
    refine = params['refine']
    domain = params['domain']

    params.setdefault('filetype', 'msh')
    params.setdefault('filename', 'out')
    params.setdefault('suppress_out', 1)
    params.setdefault('refine_custom', {}).setdefault('num_refines', 0)

    domain.setdefault('aspect_ratio', 1)
    domain.setdefault('upper_aspect_ratio', 1)
    domain.setdefault('aspect_distance', 0)
    domain.setdefault('inflow_angle', 0)
    domain.setdefault('type', 'box')

    refine.setdefault('global_scale', 1)
    refine.setdefault('turbine', {}).setdefault('num_turbines', 0)
    refine.setdefault('turbine', {}).setdefault('type', 'simple')

    refine.setdefault('farm', {}).setdefault('length_scale', params['refine']['background_length_scale'])
    refine.setdefault('farm', {}).setdefault('threshold_distance', 0)
    refine.setdefault('farm', {}).setdefault('type', 'none')

def verifyYAML(params):
    err = 0
    suppress = params['suppress_out']
    if not suppress:
        print("***----------------------------------------***")
        print("Validating YAML file.")
    for key in params:
        if key not in ['refine', 'domain', 'filetype', 'filename', 'refine_custom', 'suppress_out']:
            print("AeroMesh: Unknown field: " + key)
            err = 1
    domainChecks = params['domain']
    refineChecks = params['refine']
    customChecks = params['refine_custom']
    for key in domainChecks:
        valid = ['terrain_path', 'x_range', 'y_range', 'z_range', 'aspect_ratio', 'upper_aspect_ratio',
                 'aspect_distance', 'dimension', 'inflow_angle', 'type', 'center', 'radius']
        if key not in valid:
            print("AeroMesh: Unknown field: " + key)
            err = 1
    for key in refineChecks:
        valid = ['turbine', 'background_length_scale', 'farm', 'global_scale']
        if key not in valid:
            print("AeroMesh: Unknown field: " + key)
            err = 1
    turbineChecks = params['refine']['turbine']
    for key in turbineChecks:
        validNums = [i for i in range(1, turbineChecks['num_turbines'] + 1)]
        validParams = ['num_turbines', 'length_scale', 'threshold_upstream_distance', 'threshold_downstream_distance',
                       'threshold_rotor_distance', 'type']
        if key in validNums:
            validSubkeys = ['x', 'y', 'HH']
            for subkey in turbineChecks[key]:
                if subkey not in validSubkeys:
                    print("AeroMesh: Unknown field: " + str(key))
                    err = 1
        elif key not in validParams:
            print("AeroMesh: Unknown field: " + str(key))
            err = 1
    for key in customChecks:
        validNums = [i for i in range(1, customChecks['num_refines'] + 1)]
        if key == 'num_refines':
            continue
        elif key in validNums:
            validSubkeys = ['type', 'x_range', 'y_range', 'radius', 'length_scale', 'z_range']
            for subkey in customChecks[key]:
                if subkey not in validSubkeys:
                    print("AeroMesh: Unknown  refine_custom[" + str(key) + "] field: " + str(subkey))
                    err = 1
        else:
            print("AeroMesh: Unknown  refine_custom field: " + str(key))
            err = 1
    farmChecks = params['refine']['farm']
    for key in farmChecks:
        valid = ['length_scale', 'threshold_distance', 'type']
        if key not in valid:
            print("AeroMesh: Unknown field: " + key)
            err = 1
    if err == 0:
        if not suppress:
            print("YAML validated successfully.")
            print("***----------------------------------------***")
    else:
        print("***----------------------------------------***")
        raise Exception("YAML Error: Unexpected fields specified.")

if __name__ == "__main__":
    main()


import gmsh
import math
import numpy as np

def generateTurbines(params, domain, wf):

    """
    Builds every turbine in the range [1, num_turbines].
    It also adjusts the minimum bounding regions of the wind farm to surround
    the turbines.

    :param params: The parameter dictionary.
    :type params: dict()
    :param domain: The structure representing the domain.
    :type domain: Domain
    :param wf: The structure representing the wind farm.
    :type wf: WindFarm
    :return: A list of all the GMESH fields that represent the turbines.
    :rtype: list[int]

    """

    fields = []
    nFarms = params['refine']['turbine']['num_turbines']
    lc = params['refine']['turbine']['length_scale'] 
    lcb = params['refine']['background_length_scale']
    lcf = params['refine']['farm']['length_scale']
    rotor = params['refine']['turbine']['threshold_rotor_distance']
    aspect = params['domain']['aspect_ratio']
    aspect_upper = params['domain']['upper_aspect_ratio']
    aspect_distance = params['domain']['aspect_distance']
    inflow = params['domain']['inflow_angle']
    turbineType = params['refine']['turbine']['type']

    interp = domain.interp
    for i in range(nFarms):
        turbineData = params['refine']['turbine'][i + 1]
        x = turbineData['x'] 
        y = turbineData['y']
        hh = 100
        if turbineData.get('HH') is not None:
            hh = turbineData.get('HH')
        if interp is not None:
            z = (interp(x, y) + hh) * aspect
        else:
            z = hh * aspect

        if not domain.withinDomain(x, y, z):
            raise Exception("Invalid turbine location.")
        
        if turbineType == 'wake':
            upstream = params['refine']['turbine']['threshold_upstream_distance']
            downstream = params['refine']['turbine']['threshold_downstream_distance']
            fields.extend(placeTurbineWake(x, y, z, upstream, downstream, rotor, lc, lcb, lcf, inflow, aspect, wf, domain))
        elif turbineType == 'simple':
            ground = domain.calculateGround(x, y)
            fields.extend(placeTurbineSimple(x, y, rotor, hh, lc, lcb, ground, aspect, aspect_upper, aspect_distance, wf))
        else:
            raise Exception("Invalid turbine type. Must be simple or wake.")

    return fields

def placeTurbineSimple(x, y, rotor, hh, lc, lcb, ground, aspect, upperAspect, aspectDist, wf):
    radius = rotor
    top = ground + hh + (rotor / 2)
    bottom = ground
    bottomDist = aspectDist - bottom
    topDist = top - aspectDist

    wf.updateXMax(x)
    wf.updateXMin(x)
    wf.updateYMax(y)
    wf.updateYMin(y)
    wf.updateZMax(top + (bottomDist * (aspect - 1)) - (topDist * (1 - upperAspect)))
    
    c = gmsh.model.mesh.field.add("Cylinder")
    gmsh.model.mesh.field.setNumber(c, "Radius", radius)
    gmsh.model.mesh.field.setNumber(c, "XCenter", x)
    gmsh.model.mesh.field.setNumber(c, "YCenter", y)
    gmsh.model.mesh.field.setNumber(c, "ZAxis", top + (bottomDist * (aspect - 1)) - (topDist * (1 - upperAspect)))
    gmsh.model.mesh.field.setNumber(c, "VIn", lc)
    gmsh.model.mesh.field.setNumber(c, "VOut", lcb)

    fields = [c]
    return fields
        
def placeTurbineWake(x, y, z, upstream, downstream, rotor, lc, lcb, lcf, inflow, aspect, wf, domain):

    """
    Builds a single turbine in 3D space at the target (x, y, z) triple.
    Updates the minimum bounding region representing the farm if necessary.
    Adjusts the meshing region in the z-direction to account for anisotropic effects if needed.

    :param x: The x coordinate of the turbine center.
    :type x: double
    :param y: The y coordinate of the turbine center.
    :type y: double
    :param z: The z coordinate of the turbine center.
    :type z: double
    :param upstream: The extension of the wake in the negative direction.
    :type upstream: double
    :param downstream: The extension of the wake in the positive direction.
    :type downstream: double
    :param rotor: The radius of the rotor.
    :type rotor: double
    :param lc: The turbine level meshing restriction.
    :type lc: double
    :param lcb: The domain level meshing restriction.
    :type lcb: double
    :param lcf: The farm level meshing restriction.
    :type lcf: double
    :param inflow: The incoming wind direction.
    :type inflow: double
    :param aspect: The aspect ratio.
    :type aspect: int
    :param wf: The structure representing the wind farm.
    :type wf: WindFfarm
    :param domain: The structure representing the domain.
    :type domain: Domain
    :return: A list of GMESH fields that define the rotor's meshing region plus any anisotropic meshing in the z-direction.
    :rtype: list[int]

    """

    ###
    # Initialize proper data structures.
    ###

    increment = rotor / 2
    downPoints = math.ceil(downstream / increment)
    upPoints = math.ceil(upstream / increment)
    turbine = [gmsh.model.geo.addPoint(x, y, z)]
    anisoPoints = []

    ###
    # These loops build the body of the turbine.
    ###
    for i in range(1, downPoints + 1):
        turbine.append(gmsh.model.geo.addPoint(x + increment * i, y, z))
   
    for i in range(1, upPoints + 1):
        turbine.append(gmsh.model.geo.addPoint(x - increment * i, y, z))
    
    
    if aspect == 1:
        aspect = 0 # Locally turns off anisotropy loops.

    ###
    # These loops add the points needed to maintain the proper aspect ratio.
    ###
    for i in range(1, math.floor(aspect + 1)):
        level = []
        if z + (rotor * i) < domain.z_range[1]: 
            gmsh.model.geo.addPoint(x, y, z + (rotor * i))
        if z - (rotor * i) > domain.calculateGround(x, y):
            gmsh.model.geo.addPoint(x, y, z - (rotor * i))
        for j in range(1, downPoints + 1):
            if z + (rotor * i) < domain.z_range[1]: 
                level.append(gmsh.model.geo.addPoint(x + increment * j, y, z + (rotor * i)))
            if z - (rotor * i) > domain.calculateGround(x + increment * j, y):
                level.append(gmsh.model.geo.addPoint(x + increment * j, y, z - (rotor * i)))

        for j in range(1, upPoints + 1):
            if z + (rotor * i) < domain.z_range[1]: 
                level.append(gmsh.model.geo.addPoint(x - increment * j, y, z + (rotor * i)))
            if z - (rotor * i) > domain.calculateGround(x - increment * j, y):
                level.append(gmsh.model.geo.addPoint(x - increment * j, y, z - (rotor * i)))
        
        anisoPoints.append(level)

    ###
    # Embed all the points into the proper surfaces and volumes.
    ###

    turbineTags = list(tag for tag in zip([0] * len(turbine), turbine))
    gmsh.model.geo.rotate(turbineTags, x, y, z, 0, 0, 1, inflow)
    gmsh.model.geo.synchronize()

    for point in turbine:
        coords = gmsh.model.getValue(0, point, [])
        if domain.withinDomain(coords[0], coords[1], coords[2]):
            wf.updateXMax(coords[0])
            wf.updateXMin(coords[0])
            wf.updateYMax(coords[1])
            wf.updateYMin(coords[1])
            wf.updateZMax(coords[2])
        else:
            gmsh.model.geo.remove([(0, point)])
            turbine.remove(point)
    gmsh.model.geo.synchronize()

    for level in anisoPoints:
        levelTags = list(tag for tag in zip([0] * len(level), level))
        gmsh.model.geo.rotate(levelTags, x, y, z, 0, 0, 1, inflow)

        gmsh.model.geo.synchronize()

        for point in level:
            coords = gmsh.model.getValue(0, point, [])
            if domain.withinDomain(coords[0], coords[1], coords[2]):
                wf.updateXMax(coords[0])
                wf.updateXMin(coords[0])
                wf.updateYMax(coords[1])
                wf.updateYMin(coords[1])
                wf.updateZMax(coords[2])
            else:
                gmsh.model.geo.remove([(0, point)])
                level.remove(point)

        gmsh.model.geo.synchronize()

    gmsh.model.geo.synchronize()

    ###
    # Initialize the distance fields.
    ###
    f = gmsh.model.mesh.field.add("Distance")
    gmsh.model.mesh.field.setNumbers(f, "PointsList", turbine)

    t = gmsh.model.mesh.field.add("Threshold")
    gmsh.model.mesh.field.setNumber(t, "InField", f)
    gmsh.model.mesh.field.setNumber(t, "SizeMin", lc)
    gmsh.model.mesh.field.setNumber(t, "SizeMax", lcb)
    gmsh.model.mesh.field.setNumber(t, "DistMin", rotor)
    gmsh.model.mesh.field.setNumber(t, "DistMax", rotor + 0.5 * (lc + lcf) * 4)

    fields = [t]

    anisoScale = (aspect - 1) / 2
    a = rotor
    b = rotor + (rotor * 2 * anisoScale)

    for i, level in enumerate(anisoPoints):
        fa = gmsh.model.mesh.field.add("Distance")
        gmsh.model.mesh.field.setNumbers(fa, "PointsList", level)

        zScaled = rotor * (i + 1) #Shift so that z is centered at 0.
        
        anisoRadius = calcEllipse(a, b, zScaled)
    
        ta = gmsh.model.mesh.field.add("Threshold")
        gmsh.model.mesh.field.setNumber(ta, "InField", fa)
        gmsh.model.mesh.field.setNumber(ta, "SizeMin", lc)
        gmsh.model.mesh.field.setNumber(ta, "SizeMax", lcb)
        gmsh.model.mesh.field.setNumber(ta, "DistMin", anisoRadius)
        gmsh.model.mesh.field.setNumber(ta, "DistMax", anisoRadius + 0.5 * (lc + lcf) * 4)

        fields.append(ta)

    return fields

def anisotropyScale(params):

    """
    Compresses the mesh nodes to generate the anisotropic effects.

    :param params: The parameter dictionary.
    :type params: dict()
    """

    aspect = params['domain']['aspect_ratio']
    upper_aspect = params['domain']['upper_aspect_ratio']
    z_range = params['domain']['z_range']
    base, height = z_range[0], z_range[1]
    dist = params['domain']['aspect_distance']

    if aspect == 1:
        return

    data = gmsh.model.mesh.getNodes()
    tags = data[0]
    coords = data[1].reshape(-1, 3)

    for node in zip(tags, coords):
        tag = node[0]
        coord = node[1]
        if coord[2] <= dist * aspect:
            coord[2] /= aspect
        else:
            coord[2] -= (dist * (aspect))
            coord[2] /= ((height - dist) * upper_aspect)
            coord[2] *= (height - dist)
            coord[2] += dist
        
        gmsh.model.mesh.setNode(tag, coord, [])

def calcEllipse(a, b, z):

    """
    Calculates the distance to the border of the radius of the
    ellipse created by any anisotropic effects.

    :param a: The ellipse minor axis.
    :type a: double
    :param b: The ellipse major axis.
    :type b: double
    :param z: The elevation of the ellipse formed by the turbine, scaled so the center is at 0.
    :type z: double
    :return: The distance to the border of the anisotropic ellipse.
    :rtype: double
    """

    term = (z ** 2) / (b ** 2)
    delta = 1 - term
    delta *= (a ** 2)
    return math.sqrt(delta)

def refineFarm3D(params, wf):

    """
    Initializes a 'Box' or 'Cylinder' field that sets points within the minimum bounding regoin
    surrounding the farm to the farm's meshing constraint.

    :param params: The parameter dictionary.
    :type params: dict()
    :param wf: The structure representing the wind farm.
    :type wf: WindFarm
    :return: The GMESH field that represents the box.
    :rtype: int
    """

    lc = params['refine']['farm']['length_scale']
    jitter = params['refine']['farm']['threshold_distance']
    lcb = params['refine']['background_length_scale']
    farmType = params['refine']['farm']['type']
    
    if farmType == 'box':
        wf.adjustDistance(jitter)
        b = gmsh.model.mesh.field.add("Box")
        gmsh.model.mesh.field.setNumber(b, "XMin", wf.x_range[0])
        gmsh.model.mesh.field.setNumber(b, "XMax", wf.x_range[1])
        gmsh.model.mesh.field.setNumber(b, "YMin", wf.y_range[0])
        gmsh.model.mesh.field.setNumber(b, "YMax", wf.y_range[1])
        gmsh.model.mesh.field.setNumber(b, "ZMin", 0)
        gmsh.model.mesh.field.setNumber(b, "ZMax", wf.zMax)
        gmsh.model.mesh.field.setNumber(b, "VIn", lc)
        gmsh.model.mesh.field.setNumber(b, "VOut", lcb)

        return b
    elif farmType == 'cylinder':
        wf.adjustDistance(jitter)
        centerX = (wf.x_range[0] + wf.x_range[1]) / 2
        centerY = (wf.y_range[0] + wf.y_range[1]) / 2

        center = np.array([centerX, centerY])
        corner = np.array([wf.x_range[1], wf.y_range[1]])
        radius = np.linalg.norm(center - corner)

        c = gmsh.model.mesh.field.add("Cylinder")
        gmsh.model.mesh.field.setNumber(c, "Radius", radius)
        gmsh.model.mesh.field.setNumber(c, "VIn", lc)
        gmsh.model.mesh.field.setNumber(c, "VOut", lcb)
        gmsh.model.mesh.field.setNumber(c, "ZAxis", wf.zMax)
        gmsh.model.mesh.field.setNumber(c, "XCenter", centerX)
        gmsh.model.mesh.field.setNumber(c, "YCenter", centerY)

        return c
    else:
        return None
    
def cylinderTerrainAdjustment(domain, params):

    z_range = params['domain']['z_range']
    gradient = lambda z, h : (h - z) / h
    base, height = z_range[0], z_range[1]

    nodes = gmsh.model.mesh.getNodes()
    tags = nodes[0]
    coords = nodes[1].reshape(-1, 3)
    for tag, coord in zip(tags, coords):
        x, y, z = coord[0], coord[1], coord[2]
        interpolation = domain.interp(x, y)
        coord[2] = (interpolation * gradient(z, height)) + z

        gmsh.model.mesh.setNode(tag, coord, [])

    gmsh.model.geo.synchronize()
    gmsh.model.mesh.removeDuplicateElements()
    gmsh.model.mesh.removeDuplicateNodes()


import pathlib
import sys
sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())

project = 'AeroMesh'
copyright = '2024, National Renewable Energy Laboratory'
author = 'National Renewable Energy Laboratory'
release = 'v0.1'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_rtd_theme'
    ]

templates_path = ['_templates']
exclude_patterns = []


html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import gmsh
from aeromesh.geometry.functions3D import *
from aeromesh.structs.structures import WindFarm, Domain
import pytest
from aeromesh.terrain.terrain import buildTerrainDefault

###
# Fixtures
###
@pytest.fixture
def domain():
    d = Domain()
    d.setDomain([-1200, 1200], [-1200, 1200], [0, 1000])
    return d

@pytest.fixture
def params():
    params = dict()
    params['domain'] = {
        'x_range': [-1200, 1200],
        'y_range': [-1200, 1200],
        'z_range': [0, 1000],
        'aspect_ratio': 1,
        'upper_aspect_ratio': 1,
        'aspect_distance': 0,
        'inflow_angle': 0
    }
    params['refine'] = {
        'background_length_scale': 100,
        'turbine': {
            'type': 'wake'
        },
        'farm': {}
    }
    return params


###
# Unit Tests
###
def test_turbinegen_isolated(domain, params):
    gmsh.model.add("3D Turbine Generation Test")
    
    sl = buildTerrainDefault(params, domain)
    gmsh.model.geo.addVolume([sl], tag=999)

    wf = WindFarm()

    isotropic_test = placeTurbineWake(500, 500, 100, 240, 300, 
                                  100, 30, 100, 50, 0, 1, wf, domain)
    
    anisotropic_test = placeTurbineWake(-700, -700, 100, 240, 300, 
                                  100, 30, 100, 50, 0, 3, wf, domain)
    
    gmsh.model.remove()

    assert len(isotropic_test) == 1
    
    assert len(anisotropic_test) == 4

def test_turbinegen_full(domain, params):
    gmsh.model.add("3D Turbine Generation Test (Full)")
    wf = WindFarm()

    sl = buildTerrainDefault(params, domain)
    gmsh.model.geo.addVolume([sl], tag=999)

    params['refine']['turbine']['num_turbines'] = 2
    params['refine']['turbine']['length_scale'] = 30
    params['refine']['background_length_scale'] = 100
    params['refine']['farm']['length_scale'] = 70
    params['refine']['turbine']['threshold_upstream_distance'] = 240
    params['refine']['turbine']['threshold_downstream_distance'] = 300
    params['refine']['turbine']['threshold_rotor_distance'] = 100
    params['domain']['aspect_ratio'] = 2

    params['refine']['turbine'][1] = {
        'x': 500,
        'y': 800,
        'wake': 1
    }
    params['refine']['turbine'][2] = {
        'x': -1000,
        'y': 300,
        'wake': 0
    }

    fields_test = generateTurbines(params, domain, wf)
    gmsh.model.remove()

    assert len(fields_test) == 6

def test_anisotropy(domain, params):
    gmsh.model.add("3D Anisotropy Test")

    params['domain']['aspect_ratio'] = 2
    params['domain']['aspect_distance'] = 400

    sl = buildTerrainDefault(params, domain)
    gmsh.model.geo.addVolume([sl], tag=999)

    anisotropyScale(params)

    data = gmsh.model.mesh.getNodes()
    tags = data[0]
    coords = data[1].reshape(-1, 3)

    gmsh.model.remove()

    for node in zip(tags, coords):
        coord = node[1]
        assert coord[2] <= params['domain']['z_range'][1]



####
## Helper classes
####

class Domain():
    def __init__(self):
        self.x_range = None
        self.y_range = None
        self.z_range = None
        self.interp = None
        self.radius = None
        self.center = None

    def setDomain(self, x_range=None, y_range=None, height=[0, 1], radius=None, center=None):
        self.x_range = x_range
        self.y_range = y_range
        self.z_range = height
        self.radius = radius
        self.center = center

        if self.z_range[0] < 0:
            raise Exception("Invalid Height: Minimum height less than 0.")

    def setInterp(self, interp):
        self.interp = interp

    def calculateGround(self, x, y):
        if self.interp:
            return self.interp(x, y) + self.z_range[0]
        return self.z_range[0]

    def withinDomain(self, x, y, z=0):
        if self.x_range and self.y_range:
            if x < self.x_range[0] or x > self.x_range[1]:
                return False
            if y < self.y_range[0] or y > self.y_range[1]:
                return False
            if z > self.z_range[1] or z < self.z_range[0]:
                return False
            if self.interp and z < self.interp(x, y):
                return False
            return True
        elif self.radius and self.center:
            inCircle = lambda x, y, h, k, r: (x - h) ** 2 + (y - k) ** 2 <= r ** 2
            if not inCircle(x, y, self.center[0], self.center[1], self.radius):
                return False
            if z > self.z_range[1] or z < self.z_range[0]:
                return False
            if self.interp and z < self.interp(x, y) + self.z_range[0]:
                return False
            return True
        else:
            raise Exception("Improperly defined domain.")

class WindFarm():
    def __init__(self):
        self.zMax = 0
        self.y_range = [9999999, -9999999]
        self.x_range = [9999999, -9999999]

    def adjustDistance(self, distance):
        self.zMax += distance
        self.y_range[0] -= distance
        self.y_range[1] += distance
        self.x_range[0] -= distance
        self.x_range[1] += distance

    def updateXMax(self, x):
        if x > self.x_range[1]:
            self.x_range[1] = x
    
    def updateXMin(self, x):
        if x < self.x_range[0]:
            self.x_range[0] = x

    def updateYMin(self, y):
        if y < self.y_range[0]:
            self.y_range[0] = y
    
    def updateYMax(self, y):
        if y > self.y_range[1]:
            self.y_range[1] = y

    def updateZMax(self, z):
        if z > self.zMax:
            self.zMax = z


import gmsh
from aeromesh.terrain.terrain import buildTerrain2D, buildTerrainDefault, buildTerrainFromFile
from aeromesh.structs.structures import Domain

def test_terrain3D():
    gmsh.model.add("3D Terrain Test")
    params = dict()

    params['domain'] = {
        'terrain_path': './tests/infiles/skew_terrain.txt',
        'x_range': [-1200, 1200],
        'y_range': [-1200, 1200],
        'z_range': [0, 1000],
        'aspect_ratio': 1,
        'upper_aspect_ratio': 1,
        'aspect_distance': 0
    }

    params['refine'] = {
        'background_length_scale': 100
    }

    buildTerrainFromFile(params, Domain())
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(3)

    entities2D = gmsh.model.getEntities(dim=2)
    entities3D = gmsh.model.getEntities(dim=3)

    gmsh.model.remove()
    
    assert len(entities2D) == 6
    assert len(entities3D) == 0

def test_terrain3D_default():
    gmsh.model.add("3D Default Terrain Test")
    params = dict()

    params['domain'] = {
        'x_range': [-1200, 1200],
        'y_range': [-1200, 1200],
        'z_range': [0, 1000],
        'aspect_ratio': 1,
        'upper_aspect_ratio': 1,
        'aspect_distance': 0
    }

    params['refine'] = {
        'background_length_scale': 100
    }

    buildTerrainDefault(params, Domain())
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(3)

    entities2D = gmsh.model.getEntities(dim=2)
    entities3D = gmsh.model.getEntities(dim=3)
    gmsh.model.remove()

    assert len(entities2D) == 6
    assert len(entities3D) == 0

def test_terrain2D():
    gmsh.model.add("2D Terrain Test")
    params = dict()

    d = Domain()
    d.setDomain([-1200, 1200], [-1200, 1200])

    params['domain'] = {
        'x_range': [-1200, 1200],
        'y_range': [-1200, 1200],
    }

    params['refine'] = {
        'background_length_scale': 100
    }

    struct = buildTerrain2D(params, d)
    gmsh.model.geo.addPlaneSurface([struct])
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(2)

    entities1D = gmsh.model.getEntities(dim=1)
    entities2D = gmsh.model.getEntities(dim=2)
    gmsh.model.remove()

    assert len(entities2D) == 1
    assert len(entities1D) == 4